import numpy as np

# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).
# [2] Erlingsson, Pihur, and Korolova (2014) "RAPPOR: Randomized aggregatable privacy-preserving ordinal response" (ACM CCS).
# [3] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).

def L_SUE_Client(input_data, k, eps_perm, eps_1):

    """
    Longitudinal SUE (L-SUE) [1] protocol, a.k.a. Basic RAPPOR [2], that uses Symmetric parameters (p+q=1) [3] for both first and second rounds of sanitization.

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: sanitized value.
    """

    # SUE parameters for round 1
    p1 = np.exp(eps_perm / 2) / (np.exp(eps_perm / 2) + 1)
    q1 = 1 - p1

    # SUE parameters for round 2
    p2 = - (np.sqrt((4 * np.exp(7 * eps_perm / 2) - 4 * np.exp(5 * eps_perm / 2) - 4 * np.exp(
        3 * eps_perm / 2) + 4 * np.exp(eps_perm / 2) + np.exp(4 * eps_perm) + 4 * np.exp(3 * eps_perm) - 10 * np.exp(
        2 * eps_perm) + 4 * np.exp(eps_perm) + 1) * np.exp(eps_1)) * (np.exp(eps_1) - 1) * (
                        np.exp(eps_perm) - 1) ** 2 - (
                        np.exp(eps_1) - np.exp(2 * eps_perm) + 2 * np.exp(eps_perm) - 2 * np.exp(
                    eps_1 + eps_perm) + np.exp(eps_1 + 2 * eps_perm) - 1) * (
                        np.exp(3 * eps_perm / 2) - np.exp(eps_perm / 2) + np.exp(eps_perm) - np.exp(
                    eps_1 + eps_perm / 2) - np.exp(eps_1 + eps_perm) + np.exp(eps_1 + 3 * eps_perm / 2) + np.exp(
                    eps_1 + 2 * eps_perm) - 1)) / ((np.exp(eps_1) - 1) * (np.exp(eps_perm) - 1) ** 2 * (
                np.exp(eps_1) - np.exp(2 * eps_perm) + 2 * np.exp(eps_perm) - 2 * np.exp(eps_1 + eps_perm) + np.exp(
            eps_1 + 2 * eps_perm) - 1))
    q2 = 1 - p2

    if (np.array([p1, q1, p2, q2]) >= 0).all():
        pass
    else:
        raise ValueError('Probabilities are negative.')

    # Unary encoding
    input_ue_data = np.zeros(k)
    if input_data != None:
        input_ue_data[input_data] = 1

    # First round of sanitization (permanent memoization) with SUE using user's input_ue_data
    first_sanitization = np.zeros(k)
    for ind in range(k):
        if input_ue_data[ind] != 1:
            rnd = np.random.random()
            if rnd <= q1:
                first_sanitization[ind] = 1
        else:
            rnd = np.random.random()
            if rnd <= p1:
                first_sanitization[ind] = 1

    # Second round of sanitization with SUE using first_sanitization as input
    second_sanitization = np.zeros(k)
    for ind in range(k):
        if first_sanitization[ind] != 1:
            rnd = np.random.random()
            if rnd <= q2:
                second_sanitization[ind] = 1
        else:
            rnd = np.random.random()
            if rnd <= p2:
                second_sanitization[ind] = 1

    return second_sanitization

def L_SUE_Aggregator(ue_reports, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all L-SUE sanitized UE-based vectors;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation.
    """

    if len(ue_reports) == 0:
        raise ValueError('List of reports is empty.')

    # Number of reports
    n = len(ue_reports)

    # SUE parameters for round 1
    p1 = np.exp(eps_perm / 2) / (np.exp(eps_perm / 2) + 1)
    q1 = 1 - p1

    # SUE parameters for round 2
    p2 = - (np.sqrt((4 * np.exp(7 * eps_perm / 2) - 4 * np.exp(5 * eps_perm / 2) - 4 * np.exp(
        3 * eps_perm / 2) + 4 * np.exp(eps_perm / 2) + np.exp(4 * eps_perm) + 4 * np.exp(3 * eps_perm) - 10 * np.exp(
        2 * eps_perm) + 4 * np.exp(eps_perm) + 1) * np.exp(eps_1)) * (np.exp(eps_1) - 1) * (
                        np.exp(eps_perm) - 1) ** 2 - (
                        np.exp(eps_1) - np.exp(2 * eps_perm) + 2 * np.exp(eps_perm) - 2 * np.exp(
                    eps_1 + eps_perm) + np.exp(eps_1 + 2 * eps_perm) - 1) * (
                        np.exp(3 * eps_perm / 2) - np.exp(eps_perm / 2) + np.exp(eps_perm) - np.exp(
                    eps_1 + eps_perm / 2) - np.exp(eps_1 + eps_perm) + np.exp(eps_1 + 3 * eps_perm / 2) + np.exp(
                    eps_1 + 2 * eps_perm) - 1)) / ((np.exp(eps_1) - 1) * (np.exp(eps_perm) - 1) ** 2 * (
                np.exp(eps_1) - np.exp(2 * eps_perm) + 2 * np.exp(eps_perm) - 2 * np.exp(eps_1 + eps_perm) + np.exp(
            eps_1 + 2 * eps_perm) - 1))
    q2 = 1 - p2

    if (np.array([p1, q1, p2, q2]) >= 0).all():
        pass
    else:
        raise ValueError('Probabilities are negative.')

    # Ensure non-negativity of estimated frequency
    est_freq = ((sum(ue_reports) - n * q1 * (p2 - q2) - n * q2) / (n * (p1 - q1) * (p2 - q2))).clip(0)

    # Re-normalized estimated frequency
    norm_est_freq = est_freq / sum(est_freq)

    return norm_est_freq



import numpy as np
from pure_frequency_oracles.GRR import GRR_Client, GRR_Aggregator
from pure_frequency_oracles.UE import UE_Client, UE_Aggregator
from pure_frequency_oracles.Variance_PURE import VAR_Pure

# [1] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).

def ADP_Client(input_data, k, epsilon, optimal=True):

    """
    Adaptive (ADP) protocol that minimizes variance value (i.e., either GRR or OUE from [1]).

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: sanitized value or UE vector.
    """

    if epsilon is not None or k is not None:

        # GRR parameters
        p_grr = np.exp(epsilon) / (np.exp(epsilon) + k - 1)
        q_grr = (1 - p_grr) / (k - 1)

        # Symmetric parameters (p+q = 1)
        p_ue = np.exp(epsilon / 2) / (np.exp(epsilon / 2) + 1)
        q_ue = 1 - p_ue

        # Optimized parameters
        if optimal:
            p_ue = 1 / 2
            q_ue = 1 / (np.exp(epsilon) + 1)

        # Variance values
        var_grr = VAR_Pure(p_grr, q_grr)
        var_ue = VAR_Pure(p_ue, q_ue)

        # Adaptive protocol
        if var_grr <= var_ue:

            return GRR_Client(input_data, k, epsilon)
        else:

            return UE_Client(input_data, k, epsilon, optimal)

    else:
        raise ValueError('k (int) and epsilon (float) need a numerical value.')


def ADP_Aggregator(reports, k, epsilon, optimal=True):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all GRR sanitized values or UE-based vectors;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: normalized frequency (histogram) estimation.
    """

    if len(reports) == 0:

        raise ValueError('List of reports is empty.')

    else:

        if epsilon is not None or k is not None:

            # GRR parameters
            p_grr = np.exp(epsilon) / (np.exp(epsilon) + k - 1)
            q_grr = (1 - p_grr) / (k - 1)

            # Symmetric parameters (p+q = 1)
            p_ue = np.exp(epsilon / 2) / (np.exp(epsilon / 2) + 1)
            q_ue = 1 - p_ue

            # Optimized parameters
            if optimal:
                p_ue = 1 / 2
                q_ue = 1 / (np.exp(epsilon) + 1)

            # Variance values
            var_grr = VAR_Pure(p_grr, q_grr)
            var_ue = VAR_Pure(p_ue, q_ue)

            # Adaptive estimator
            if var_grr <= var_ue:

                return GRR_Aggregator(reports, k, epsilon)
            else:

                return UE_Aggregator(reports, epsilon, optimal)

        else:
            raise ValueError('k (int) and epsilon (float) need a numerical value.')



from long_mdim_freq_est.L_SMP_Solution import SMP_L_GRR_Client, SMP_L_OUE_Client, SMP_L_OSUE_Client, SMP_L_SUE_ClientSMP_L_SOUE_Client, SMP_L_ADP_Client, SMP_L_GRR_Aggregator, SMP_L_OUE_Aggregator, SMP_L_OSUE_Aggregator, SMP_L_SUE_Aggregator, SMP_L_SOUE_Aggregator, SMP_L_ADP_Aggregator
from long_mdim_freq_est.L_SPL_Solution import SPL_L_GRR_Client, SPL_L_OUE_Client, SPL_L_OSUE_Client, SPL_L_SUE_Client, SPL_L_SOUE_Client, SPL_L_ADP_Client, SPL_L_GRR_Aggregator, SPL_L_OUE_Aggregator, SPL_L_OSUE_Aggregator, SPL_L_SUE_Aggregator, SPL_L_SOUE_Aggregator, SPL_L_ADP_Aggregator


import numpy as np
from numba import jit

# [1] Erlingsson, Pihur, and Korolova (2014) "RAPPOR: Randomized aggregatable privacy-preserving ordinal response" (ACM CCS).
# [2] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).

@jit(nopython=True)
def UE_Client(input_data, k, epsilon, optimal=True):
    """
    Unary Encoding (UE) protocol, a.k.a. Basic One-Time RAPPOR (if optimal=False) [1]

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: sanitized UE vector.
    """

    # Symmetric parameters (p+q = 1)
    p = np.exp(epsilon/2) / (np.exp(epsilon/2) + 1)
    q = 1 - p

    # Optimized parameters
    if optimal:
        p = 1 / 2
        q = 1 / (np.exp(epsilon) + 1)
        
    # Unary encoding
    input_ue_data = np.zeros(k)
    if input_data != None:
        input_ue_data[input_data] = 1

    # Initializing a zero-vector
    sanitized_vec = np.zeros(k)

    # UE perturbation function
    for ind in range(k):
        if input_ue_data[ind] != 1:
            rnd = np.random.random()
            if rnd <= q:
                sanitized_vec[ind] = 1
        else:
            rnd = np.random.random()
            if rnd <= p:
                sanitized_vec[ind] = 1
    return sanitized_vec
        
def UE_Aggregator(reports, epsilon, optimal=True):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all UE-based sanitized vectors;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: normalized frequency (histogram) estimation.
    """

    if len(reports) == 0: 
        
        raise ValueError('List of reports is empty.')
        
    else:

        if epsilon is not None:

            # Number of reports
            n = len(reports)

            # Symmetric parameters (p+q = 1)
            p = np.exp(epsilon/2) / (np.exp(epsilon/2) + 1)
            q = 1 - p

            # Optimized parameters
            if optimal:
                p = 1 / 2
                q = 1 / (np.exp(epsilon) + 1)

            # Ensure non-negativity of estimated frequency
            est_freq = np.array((sum(reports) - q * n) / (p-q)).clip(0)

            # Re-normalized estimated frequency
            norm_est_freq = est_freq / sum(est_freq)

            return norm_est_freq

        else:
            raise ValueError('epsilon (float) needs a numerical value.')



import numpy as np
from numba import jit

# [1] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).
# [2] Kairouz, Bonawitz, and Ramage (2016) "Discrete distribution estimation under local privacy" (ICML)

@jit(nopython=True)
def GRR_Client(input_data, k, epsilon):
    """
    Generalized Randomized Response (GRR) protocol, a.k.a., direct encoding [1] or k-RR [2].

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :return: sanitized value.
    """

    if epsilon is not None or k is not None:
        
        # GRR parameters
        p = np.exp(epsilon) / (np.exp(epsilon) + k - 1)

        # Mapping domain size k to the range [0, ..., k-1]
        domain = np.arange(k) 
        
        # GRR perturbation function
        rnd = np.random.random()
        if rnd <= p:
            return input_data

        else:
            return np.random.choice(domain[domain != input_data])

    else:
        raise ValueError('k (int) and epsilon (float) need a numerical value.')


def GRR_Aggregator(reports, k, epsilon):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all GRR-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :return: normalized frequency (histogram) estimation.
    """

    if len(reports) == 0:

        raise ValueError('List of reports is empty.')
        
    else:

        if epsilon is not None or k is not None:
            
            # Number of reports
            n = len(reports)
            
            # GRR parameters
            p = np.exp(epsilon) / (np.exp(epsilon) + k - 1)
            q = (1 - p) / (k - 1)

            # Count how many times each value has been reported
            count_report = np.zeros(k)
            for rep in reports: 
                count_report[rep] += 1

            # Ensure non-negativity of estimated frequency
            est_freq = np.array((count_report - n*q) / (p-q)).clip(0)

            # Re-normalized estimated frequency
            norm_est_freq = est_freq / sum(est_freq) 

            return norm_est_freq

        else:
            raise ValueError('k (int) and epsilon (float) need a numerical value.')



import numpy as np
from pure_frequency_oracles.GRR import GRR_Client

# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).
# [2] Kairouz, Bonawitz, and Ramage (2016) "Discrete distribution estimation under local privacy" (ICML)

def L_GRR_Client(input_data, k, eps_perm, eps_1):

    """
    Longitudinal GRR (L-GRR) [1] protocol that chaines GRR [2] for both first and second rounds of sanitization.

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: sanitized value.
    """

    # GRR parameters for round 1
    p1 = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
    q1 = (1 - p1) / (k - 1)

    # GRR parameters for round 2
    p2 = (q1 - np.exp(eps_1) * p1) / ((-p1 * np.exp(eps_1)) + k*q1*np.exp(eps_1) - q1*np.exp(eps_1) - p1*(k-1)+q1)
    q2 = (1 - p2) / (k-1)
    
    if (np.array([p1, q1, p2, q2]) >= 0).all():
        pass
    else: 
        raise ValueError('Probabilities are negative.')

    # Get epsilon of second round of sanitization
    eps_sec_round = np.log(p2 / q2)

    # First round of sanitization (permanent memoization) with GRR using user's input_data
    first_sanitization = GRR_Client(input_data, k, eps_perm)

    # Second round of sanitization with GRR using first_sanitization as input
    second_sanitization = GRR_Client(first_sanitization, k, eps_sec_round)
    
    return second_sanitization

def L_GRR_Aggregator(reports, k, eps_perm, eps_1):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all L-GRR sanitized values;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation.
    """

    if len(reports) == 0:
        raise ValueError('List of reports is empty.')

    # Number of reports
    n = len(reports)
                
    # GRR parameters for round 1
    p1 = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
    q1 = (1 - p1) / (k - 1)

    # GRR parameters for round 2
    p2 = (q1 - np.exp(eps_1) * p1) / ((-p1 * np.exp(eps_1)) + k*q1*np.exp(eps_1) - q1*np.exp(eps_1) - p1*(k-1)+q1)
    q2 = (1 - p2) / (k-1)
    
    if (np.array([p1, q1, p2, q2]) >= 0).all():
        pass
    else: 
        raise ValueError('Probabilities are negative.')

    # Count how many times each value has been reported
    count_report = np.zeros(k)            
    for rep in reports:
        count_report[rep] += 1

    # Ensure non-negativity of estimated frequency
    est_freq = ((count_report - n*q1*(p2-q2) - n*q2) / (n*(p1-q1)*(p2-q2))).clip(0)

    # Re-normalized estimated frequency
    norm_est_freq = est_freq / sum(est_freq)
    
    return norm_est_freq



from mdim_freq_est.RSpFD_solution import RSpFD_GRR_Client, RSpFD_UE_zero_Client, RSpFD_UE_rnd_Client, RSpFD_ADP_Client, RSpFD_GRR_Aggregator, RSpFD_UE_zero_Aggregator, RSpFD_UE_rnd_Aggregator, RSpFD_ADP_Aggregator
from mdim_freq_est.SMP_solution import SMP_GRR_Client, SMP_UE_Client, SMP_ADP_Client, SMP_GRR_Aggregator, SMP_UE_Aggregator, SMP_ADP_Aggregator, 
from mdim_freq_est.SPL_solution import SPL_GRR_Client, SPL_UE_Client, SPL_ADP_Client, SPL_GRR_Aggregator, SPL_UE_Aggregator, SPL_ADP_Aggregator
from mdim_freq_est.Variance_RSpFD import VAR_RSpFD_GRR, VAR_RSpFD_UE_zero, VAR_RSpFD_UE_rnd


import numpy as np
from long_freq_est.L_GRR import L_GRR_Client, L_GRR_Aggregator
from long_freq_est.L_OUE import L_OUE_Client, L_OUE_Aggregator
from long_freq_est.L_OSUE import L_OSUE_Client, L_OSUE_Aggregator
from long_freq_est.L_SUE import L_SUE_Client, L_SUE_Aggregator
from long_freq_est.L_SOUE import L_SOUE_Client, L_SOUE_Aggregator
from long_freq_est.L_ADP import L_ADP_Client, L_ADP_Aggregator

# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).
# [2] Erlingsson, Pihur, and Korolova (2014) "RAPPOR: Randomized aggregatable privacy-preserving ordinal response" (ACM CCS).

def SPL_L_GRR_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Splitting (SPL) the privacy budget and using L-GRR [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple of sanitized values.
    """

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Sanitization of each value with L-GRR protocol
    sanitized_tuple = []
    for idx in range(d):
        sanitized_tuple.append(L_GRR_Client(input_tuple[idx], lst_k[idx], eps_perm_spl, eps_1_spl))

    return sanitized_tuple

def SPL_L_OUE_Client(input_tuple, lst_k, d, eps_perm, eps_1):
    """
    Splitting (SPL) the privacy budget and using L-OUE [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple of sanitized UE vectors.
    """

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Sanitization of each value with L-OUE protocol
    sanitized_tuple = []
    for idx in range(d):
        sanitized_tuple.append(L_OUE_Client(input_tuple[idx], lst_k[idx], eps_perm_spl, eps_1_spl))

    return sanitized_tuple

def SPL_L_OSUE_Client(input_tuple, lst_k, d, eps_perm, eps_1):
    """
    Splitting (SPL) the privacy budget and using L-OSUE [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple of sanitized UE vectors.
    """

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Sanitization of each value with L-OSUE protocol
    sanitized_tuple = []
    for idx in range(d):
        sanitized_tuple.append(L_OSUE_Client(input_tuple[idx], lst_k[idx], eps_perm_spl, eps_1_spl))

    return sanitized_tuple

def SPL_L_SUE_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Splitting (SPL) the privacy budget and using L-SUE [1] (a.k.a. Basic RAPPOR [2]) protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple of sanitized UE vectors.
    """

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Sanitization of each value with L-SUE protocol
    sanitized_tuple = []
    for idx in range(d):
        sanitized_tuple.append(L_SUE_Client(input_tuple[idx], lst_k[idx], eps_perm_spl, eps_1_spl))

    return sanitized_tuple

def SPL_L_SOUE_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Splitting (SPL) the privacy budget and using L-SOUE [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple of sanitized UE vectors.
    """

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Sanitization of each value with L-SOUE protocol
    sanitized_tuple = []
    for idx in range(d):
        sanitized_tuple.append(L_SOUE_Client(input_tuple[idx], lst_k[idx], eps_perm_spl, eps_1_spl))

    return sanitized_tuple

def SPL_L_ADP_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Splitting (SPL) the privacy budget and using L-ADP [1] protocol (a.k.a. ALLOMFREE [1]) as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple of sanitized UE vectors.
    """

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Sanitization of each value with L-ADP protocol
    sanitized_tuple = []
    for idx in range(d):
        sanitized_tuple.append(L_ADP_Client(input_tuple[idx], lst_k[idx], eps_perm_spl, eps_1_spl))

    return sanitized_tuple

def SPL_L_GRR_Aggregator(reports_tuple, lst_k, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all sanitized tuples;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        lst_freq_est.append(L_GRR_Aggregator(reports, lst_k[idx], eps_perm_spl, eps_1_spl))

    return lst_freq_est

def SPL_L_OUE_Aggregator(reports_tuple, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all sanitized tuples;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        lst_freq_est.append(L_OUE_Aggregator(reports, eps_perm_spl, eps_1_spl))

    return lst_freq_est

def SPL_L_OSUE_Aggregator(reports_tuple, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all sanitized tuples;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        lst_freq_est.append(L_OSUE_Aggregator(reports, eps_perm_spl, eps_1_spl))

    return lst_freq_est

def SPL_L_SUE_Aggregator(reports_tuple, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all sanitized tuples;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        lst_freq_est.append(L_SUE_Aggregator(reports, eps_perm_spl, eps_1_spl))

    return lst_freq_est

def SPL_L_SOUE_Aggregator(reports_tuple, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all sanitized tuples;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        lst_freq_est.append(L_SOUE_Aggregator(reports, eps_perm_spl, eps_1_spl))

    return lst_freq_est

def SPL_L_ADP_Aggregator(reports_tuple, lst_k, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all sanitized tuples;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Splitting the privacy budget over the number of attributes d
    eps_perm_spl = eps_perm / d
    eps_1_spl = eps_1 / d

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        lst_freq_est.append(L_ADP_Aggregator(reports, lst_k[idx], eps_perm_spl, eps_1_spl))

    return lst_freq_est



import numpy as np
from pure_frequency_oracles.GRR import GRR_Client, GRR_Aggregator
from pure_frequency_oracles.UE import UE_Client, UE_Aggregator
from pure_frequency_oracles.ADP import ADP_Client, ADP_Aggregator

# [1] Kairouz, Bonawitz, and Ramage (2016) "Discrete distribution estimation under local privacy" (ICML)
# [2] Erlingsson, Pihur, and Korolova (2014) "RAPPOR: Randomized aggregatable privacy-preserving ordinal response" (ACM CCS).
# [3] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).

def SMP_GRR_Client(input_tuple, lst_k, d, epsilon):

    """
    Sampling (SMP) a single attribute and using Generalized Randomized Response (GRR) [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :return: tuple (sampled attribute, sanitized value).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with GRR protocol
    att_sanitized_value = (rnd_att, GRR_Client(input_tuple[rnd_att], lst_k[rnd_att], epsilon))

    return att_sanitized_value

def SMP_UE_Client(input_tuple, lst_k, d, epsilon, optimal=True):

    """
    Sampling (SMP) a single attribute and using Unary Encoding (UE) protocol [2] as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [3];
    :return: tuple (sampled attribute, sanitized UE vector).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with UE protocol
    att_sanitized_value = (rnd_att, UE_Client(input_tuple[rnd_att], lst_k[rnd_att], epsilon, optimal))
    
    return att_sanitized_value

def SMP_ADP_Client(input_tuple, lst_k, d, epsilon, optimal=True):

    """
    Sampling (SMP) a single attribute and using Adaptive (ADP) protocol (i.e., GRR or UE) as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [3];
    :return: tuple (sampled attribute, sanitized value or UE vector).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with ADP protocol
    att_sanitized_value = (rnd_att, ADP_Client(input_tuple[rnd_att], lst_k[rnd_att], epsilon, optimal))

    return att_sanitized_value

def SMP_GRR_Aggregator(reports_tuple, lst_k, d, epsilon):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all tuples (sampled attribute, sanitized value);
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_tuple) == 0:

        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att:[] for att in range(d)}
    for val in reports_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(GRR_Aggregator(dic_rep_smp[idx], lst_k[idx], epsilon))

    return lst_freq_est

def SMP_UE_Aggregator(reports_tuple, d, epsilon, optimal=True):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all tuples (sampled attribute, sanitized UE vector);
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [3];
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_tuple) == 0:

        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att:[] for att in range(d)}
    for val in reports_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(UE_Aggregator(dic_rep_smp[idx], epsilon, optimal))

    return lst_freq_est

def SMP_ADP_Aggregator(reports_tuple, lst_k, d, epsilon, optimal=True):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all tuples (sampled attribute, sanitized value or UE vector);
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [3];
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_tuple) == 0:
        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att: [] for att in range(d)}
    for val in reports_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):

        lst_freq_est.append(ADP_Aggregator(dic_rep_smp[idx], lst_k[idx], epsilon, optimal))

    return lst_freq_est



from pure_frequency_oracles.ADP import ADP_Client, ADP_Aggregator
from pure_frequency_oracles.GRR import GRR_Client, GRR_Aggregator
from pure_frequency_oracles.UE import UE_Client, UE_Aggregator
from pure_frequency_oracles.Variance_PURE import VAR_Pure


import setuptools
from setuptools import setup, find_packages
import os

with open("README.md", "r") as fh:
    describe_package = fh.read()

setup(
    name='multi-freq-ldpy',
    version='0.0.5',
    license='MIT',
    author="Héber H. Arcolezi",
    author_email='hh.arcolezi@gmail.com',
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    url='https://github.com/hharcolezi/multi-freq-ldpy',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords='local-differential-privacy',
    install_requires=[
                        'numpy', 'numba'
                     ],
    description='Multiple Frequency Estimation Under Local Differential Privacy in Python',
    long_description=describe_package,
    python_requires=">=3.6",

)


from long_freq_est.L_ADP import L_ADP_Client, L_ADP_Aggregator
from long_freq_est.L_GRR import L_GRR_Client, L_GRR_Aggregator
from long_freq_est.L_OSUE import L_OSUE_Client, L_OSUE_Aggregator
from long_freq_est.L_OUE import L_OUE_Client, L_OUE_Aggregator
from long_freq_est.L_SOUE import L_SOUE_Client, L_SOUE_Aggregator
from long_freq_est.L_SUE import L_SUE_Client, L_SUE_Aggregator
from long_freq_est.Variance_LONG_PURE import VAR_Long_Pure


import numpy as np
from long_freq_est.L_GRR import L_GRR_Client, L_GRR_Aggregator
from long_freq_est.L_OSUE import L_OSUE_Client, L_OSUE_Aggregator
from long_freq_est.Variance_LONG_PURE import VAR_Long_Pure

# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).

def L_ADP_Client(input_data, k, eps_perm, eps_1):

    """
    Longitudinal Adaptive (L-ADP) protocol that minimizes variance value (i.e., either L-GRR or L-OSUE from [1], a.k.a. ALLOMFREE).

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: sanitized value or UE vector.
    """

    # GRR parameters for round 1
    p1_grr = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
    q1_grr = (1 - p1_grr) / (k - 1)

    #  GRR parameters for round 2
    p2_grr = (q1_grr - np.exp(eps_1) * p1_grr) / (
                (-p1_grr * np.exp(eps_1)) + k * q1_grr * np.exp(eps_1) - q1_grr * np.exp(eps_1) - p1_grr * (k - 1) + q1_grr)
    q2_grr = (1 - p2_grr) / (k - 1)

    if (np.array([p1_grr, q1_grr, p2_grr, q2_grr]) >= 0).all():
        pass
    else:
        raise ValueError('Probabilities are negative.')

    # OUE parameters for round 1
    p1_ue = 1 / 2
    q1_ue = 1 / (np.exp(eps_perm) + 1)

    # SUE parameters for round 2
    p2_ue = (1 - np.exp(eps_1 + eps_perm)) / (np.exp(eps_1) - np.exp(eps_perm) - np.exp(eps_1 + eps_perm) + 1)
    q2_ue = 1 - p2_ue

    if (np.array([p1_ue, q1_ue, p2_ue, q2_ue]) >= 0).all():
        pass
    else:
        raise ValueError('Probabilities are negative.')

    # Variance values of L-GRR and L-OSUE
    var_l_grr = VAR_Long_Pure(p1_grr, q1_grr, p2_grr, q2_grr)
    var_l_osue = VAR_Long_Pure(p1_ue, q1_ue, p2_ue, q2_ue)

    # Adaptive longitudinal protocol (a.k.a. ALLOMFREE in [1])
    if var_l_grr <= var_l_osue:

        return L_GRR_Client(input_data, k, eps_perm, eps_1)
    else:

        return L_OSUE_Client(input_data, k, eps_perm, eps_1)

def L_ADP_Aggregator(reports, k, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all sanitized L-GRR values or L-OSUE UE-based vectors;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation.
    """

    # GRR parameters for round 1
    p1_grr = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
    q1_grr = (1 - p1_grr) / (k - 1)

    #  GRR parameters for round 2
    p2_grr = (q1_grr - np.exp(eps_1) * p1_grr) / (
            (-p1_grr * np.exp(eps_1)) + k * q1_grr * np.exp(eps_1) - q1_grr * np.exp(eps_1) - p1_grr * (k - 1) + q1_grr)
    q2_grr = (1 - p2_grr) / (k - 1)

    if (np.array([p1_grr, q1_grr, p2_grr, q2_grr]) >= 0).all():
        pass
    else:
        raise ValueError('Probabilities are negative.')

    # OUE parameters for round 1
    p1_ue = 1 / 2
    q1_ue = 1 / (np.exp(eps_perm) + 1)

    # SUE parameters for round 2
    p2_ue = (1 - np.exp(eps_1 + eps_perm)) / (np.exp(eps_1) - np.exp(eps_perm) - np.exp(eps_1 + eps_perm) + 1)
    q2_ue = 1 - p2_ue

    if (np.array([p1_ue, q1_ue, p2_ue, q2_ue]) >= 0).all():
        pass
    else:
        raise ValueError('Probabilities are negative.')

    # Variance values of L-GRR and L-OSUE
    var_l_grr = VAR_Long_Pure(p1_grr, q1_grr, p2_grr, q2_grr)
    var_l_osue = VAR_Long_Pure(p1_ue, q1_ue, p2_ue, q2_ue)

    # Adaptive estimator
    if var_l_grr <= var_l_osue:

        return L_GRR_Aggregator(reports, k, eps_perm, eps_1)

    else:
        return L_OSUE_Aggregator(reports, eps_perm, eps_1)



import numpy as np

def LH_Client(input_data, , epsilon, optimal=True):
    
    
    
    
    
    return


def LH_Aggregator(reports, , epsilon, optimal=True):
    
    
    
    
    
    
    return
    



import numpy as np
from pure_frequency_oracles.GRR import GRR_Client, GRR_Aggregator
from pure_frequency_oracles.UE import UE_Client, UE_Aggregator
from pure_frequency_oracles.ADP import ADP_Client, ADP_Aggregator

# [1] Kairouz, Bonawitz, and Ramage (2016) "Discrete distribution estimation under local privacy" (ICML)
# [2] Erlingsson, Pihur, and Korolova (2014) "RAPPOR: Randomized aggregatable privacy-preserving ordinal response" (ACM CCS).
# [3] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).

def SPL_GRR_Client(input_tuple, lst_k, d, epsilon):

	"""
	Splitting (SPL) the privacy budget and using Generalized Randomized Response (GRR) [1] protocol as local randomizer.

	:param input_tuple: user's true tuple of values;
	:param lst_k: list of attributes' domain size;
	:param d: number of attributes;
	:param epsilon: privacy guarantee;
	:return: tuple of sanitized values.
	"""

	# Splitting the privacy budget over the number of attributes d
	eps_spl = epsilon / d

	# Sanitization of each value with GRR protocol
	sanitized_tuple = []
	for idx in range(d):
		
		sanitized_tuple.append(GRR_Client(input_tuple[idx], lst_k[idx], eps_spl))

	return sanitized_tuple

def SPL_UE_Client(input_tuple, lst_k, d, epsilon, optimal=True):
	"""
	Splitting (SPL) the privacy budget and using Unary Encoding (UE) [2] protocol as local randomizer.

	:param input_tuple: user's true tuple of values;
	:param lst_k: list of attributes' domain size;
	:param d: number of attributes;
	:param epsilon: privacy guarantee;
	:param optimal: if True, it uses the Optimized UE (OUE) protocol from [3];
	:return: tuple of sanitized UE vectors.
	"""

	# Splitting the privacy budget over the number of attributes d
	eps_spl = epsilon / d

	# Sanitization of each value with UE protocol
	sanitized_ue_tuple = []
	for idx in range(d):

		sanitized_ue_tuple.append(UE_Client(input_tuple[idx], lst_k[idx], eps_spl, optimal))

	return sanitized_ue_tuple

def SPL_ADP_Client(input_tuple, lst_k, d, epsilon, optimal=True):

	"""
	Splitting (SPL) the privacy budget and using Adaptive (ADP) protocol (i.e., GRR or UE) as local randomizer.

	:param input_tuple: user's true tuple of values;
	:param lst_k: list of attributes' domain size;
	:param d: number of attributes;
	:param epsilon: privacy guarantee;
	:param optimal: if True, it uses the Optimized UE (OUE) protocol from [3];
	:return: tuple of sanitized UE vectors.
	"""

	# Splitting the privacy budget over the number of attributes d
	eps_spl = epsilon / d

	# Sanitization of each value with ADP protocol
	sanitized_tuple = []
	for idx in range(d):

		sanitized_tuple.append(ADP_Client(input_tuple[idx], lst_k[idx], eps_spl, optimal))

	return sanitized_tuple

def SPL_GRR_Aggregator(reports_tuple, lst_k, d, epsilon):

	"""
	Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

	:param reports_tuple: list of all sanitized tuples;
	:param lst_k: list of attributes' domain size;
	:param d: number of attributes;
	:param epsilon: privacy guarantee;
	:return: normalized frequency (histogram) estimation of all d attributes.
	"""

	if len(reports_tuple) == 0:
		raise ValueError('List of reports is empty.')

	reports_tuple = np.array(reports_tuple, dtype='object')

	# Splitting the privacy budget over the number of attributes d
	eps_spl = epsilon / d

	# Estimated frequency for all d attributes
	lst_freq_est = []
	for idx in range(d):
		
		reports = reports_tuple[:, idx]
		lst_freq_est.append(GRR_Aggregator(reports, lst_k[idx], eps_spl))

	return lst_freq_est

def SPL_UE_Aggregator(reports_tuple, d, epsilon, optimal=True):

	"""
	Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

	:param reports_tuple: list of all sanitized tuples;
	:param lst_k: list of attributes' domain size;
	:param d: number of attributes;
	:param epsilon: privacy guarantee;
	:param optimal: if True, it uses the Optimized UE (OUE) protocol from [3];
	:return: normalized frequency (histogram) estimation of all d attributes.
	"""

	if len(reports_tuple) == 0:
		raise ValueError('List of reports is empty.')

	reports_tuple = np.array(reports_tuple, dtype='object')

	# Splitting the privacy budget over the number of attributes d
	eps_spl = epsilon / d

	# Estimated frequency for all d attributes
	lst_freq_est = []
	for idx in range(d):
		
		reports_ue = reports_tuple[:, idx]
		lst_freq_est.append(UE_Aggregator(reports_ue, eps_spl, optimal))

	return lst_freq_est

def SPL_ADP_Aggregator(reports_tuple, lst_k, d, epsilon, optimal=True):

	"""
	Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

	:param reports_tuple: list of all sanitized tuples;
	:param lst_k: list of attributes' domain size;
	:param d: number of attributes;
	:param epsilon: privacy guarantee;
	:param optimal: if True, it uses the Optimized UE (OUE) protocol from [3];
	:return: normalized frequency (histogram) estimation of all d attributes.
	"""

	if len(reports_tuple) == 0:
		raise ValueError('List of reports is empty.')

	reports_tuple = np.array(reports_tuple, dtype='object')

	# Splitting the privacy budget over the number of attributes d
	eps_spl = epsilon / d

	# Estimated frequency for all d attributes
	lst_freq_est = []
	for idx in range(d):

		reports = reports_tuple[:, idx]
		lst_freq_est.append(ADP_Aggregator(reports, lst_k[idx], eps_spl, optimal))

	return lst_freq_est
